#ifndef _BOARD_H
#define _BOARD_H

#include "color.h"
#include "packedBoard.h"
#include "pos2IndexTable.h"
/**
 *
 */
class Board {
public:
  static const int WIDTH=9;
  static const int HEIGHT=6;
  static const int SIZE=WIDTH*HEIGHT;
  static const int RIGHT=1;
  static const int LEFT=-RIGHT;
  static const int DOWN=WIDTH;
  static const int UP= -DOWN;
  static const int RIGHTDOWN=5;
  static const int H_START=RIGHT+DOWN;
  static const int H_SIZE=4;
  static const int H_INDEX=0;
  static const int L_START=H_START+RIGHTDOWN;
  static const int L_SIZE=3;
  static const int L_INDEX=16;
private:
  Color board[SIZE];
  void init(Board const & b){
    std::fill(&board[0],&board[SIZE],EDGE);
    for(int pos=0;pos<25;pos++){
      int i=pos2IndexTable.getIndex(pos);
      board[i]=b.board[i];
    }
  }
  int flip(int i,int offset){
    int i1;
    for(i1=i+offset;board[i1]==BLUE;i1+=offset);
    if(board[i1]!=RED) return 0;
    int ret=0;
    for(i1=i+offset;board[i1]==BLUE;i1+=offset){
      board[i1]=EMPTY;
      ret++;
    }
    return ret;
  }
public:
  Board();
  Board(const char *);
  const Color& operator[](int i) const{
    assert(0<=i && i<SIZE);
    return board[pos2IndexTable.getIndex(i)];
  }
  void init(PackedBoard v){
    int rot=v&7;
    for(int i=0;i<25;i++){
      Color c=static_cast<Color>((v>>(i*2+3))&3);
      board[pos2IndexTable.getIndex(rot,i)]=c;
    }
  }
  Color& operator[](int i){
    assert(0<=i && i<SIZE);
    return board[pos2IndexTable.getIndex(i)];
  }
  int nStones() const{
    int ret=0;
    for(int i=0;i<25;i++){
      if(board[pos2IndexTable.getIndex(i)]!=EMPTY) ret++;
    }
    return ret;
  }
  int nStones(Color c) const{
    int ret=0;
    for(int i=0;i<25;i++){
      if(board[pos2IndexTable.getIndex(i)]==c) ret++;
    }
    return ret;
  }
  void revColor(){
    for(int i=0;i<25;i++){
      int j=pos2IndexTable.getIndex(i);
      board[j]=alt(board[j]);
    }
  }
  int count(int i,int offset) const
  {
    int count=0;
    for(i+=offset;board[i]==RED;i+=offset){
      count++;
    }
    return count;
  }
  bool canWinByDrop(int i) const{
    if(board[i]!=EMPTY) return false;
    if(count(i,RIGHT)+count(i,-RIGHT)==2) return true;
    if(count(i,DOWN)+count(i,-DOWN)==2) return true;
    if(count(i,RIGHT+DOWN)+count(i,-RIGHT-DOWN)==2) return true;
    if(count(i,RIGHT-DOWN)+count(i,-RIGHT+DOWN)==2) return true;
    return false;
  }
  bool canWinByMove(int pos) const{
    int i=pos2IndexTable.getIndex(pos);
    if(board[i]!=RED) return false;
    if(canWinByDrop(i-RIGHTDOWN)) return true;
    if(canWinByDrop(i-RIGHTDOWN+RIGHT)) return true;
    if(canWinByDrop(i-RIGHTDOWN+DOWN)) return true;
    if(canWinByDrop(i-RIGHTDOWN+RIGHT+DOWN)) return true;
    return false;
  }
  bool canWinByDrop() const{
    assert(nStones()<8);
    for(int k=0;k<25;k++){
      int i=pos2IndexTable.getIndex(k);
      if(canWinByDrop(i)){
	return true;
      }
    }
    return false;
  }
  bool canWinByMove() const{
    //    assert(nStones()==8);
    for(int k=0;k<25;k++)
      if(canWinByMove(k)) return true;
    return false;
  }
  bool canWin() const{
    if(nStones(RED)<4) return canWinByDrop();
    else return canWinByMove();
  }
  PackedBoard packRev(int i) const{
    PackedBoard val=i;
    for(int j=0;j<25;j++)
      val|=static_cast<PackedBoard>(alt(board[pos2IndexTable.getIndex(i,j)]))<<(j*2+3);
    return val;
  }
  PackedBoard packRev() const{
    PackedBoard val=packRev(0);
    for(int i=1;i<8;i++)
      val=min(val,packRev(i));
    return val;
  }
  void nextDrops(vector<PackedBoard>& ret,Board &newBoard,int pos,int count);
  void nextDrops(vector<PackedBoard> & ret,int i){
    if(board[i]!=EMPTY) return;
    Board newBoard(*this);
    newBoard.board[i]=RED;
    int count=0;
    count+=newBoard.flip(i,RIGHT);
    count+=newBoard.flip(i,-RIGHT);
    count+=newBoard.flip(i,DOWN);
    count+=newBoard.flip(i,-DOWN);
    count+=newBoard.flip(i,RIGHT+DOWN);
    count+=newBoard.flip(i,-RIGHT-DOWN);
    count+=newBoard.flip(i,RIGHT-DOWN);
    count+=newBoard.flip(i,-RIGHT+DOWN);
    nextDrops(ret,newBoard,0,count);
  }
  void nextMoves(vector<PackedBoard>& nexts,int i){
    if(board[i]!=RED) return;
    board[i]=EMPTY;
    nextDrops(nexts,i-RIGHTDOWN);
    nextDrops(nexts,i-RIGHTDOWN+RIGHT);
    nextDrops(nexts,i-RIGHTDOWN+DOWN);
    nextDrops(nexts,i-RIGHTDOWN+RIGHT+DOWN);
    board[i]=RED;
  }
  void nextMoves(vector<PackedBoard>& nexts){
    for(int pos=0;pos<25;pos++)
      nextMoves(nexts,pos2IndexTable.getIndex(pos));
    if(nexts.size()==0){
      nexts.push_back(packRev());
    }
  }
  void nextDrops(vector<PackedBoard>& nexts){
    for(int pos=0;pos<25;pos++)
      nextDrops(nexts,pos2IndexTable.getIndex(pos));
  }
  void nextStates(vector<PackedBoard>& nexts){
    if(nStones()<8) nextDrops(nexts);
    else nextMoves(nexts);
  }
};

ostream& operator<<(ostream& os,Board const& b);
inline void Board::nextDrops(vector<PackedBoard>& ret,Board &newBoard,int pos,int count){
    if(count==0){
      ret.push_back(newBoard.packRev());
    }
    else if(pos<25){
      int i=pos2IndexTable.getIndex(pos);
      if(board[i]==EMPTY && newBoard.board[i]!=RED){
	newBoard.board[i]=BLUE;
	nextDrops(ret,newBoard,pos+1,count-1);
	newBoard.board[i]=EMPTY;
      }
      nextDrops(ret,newBoard,pos+1,count);
    }
  }

#endif /* _BOARD_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:

